@@ -5,7 +5,7 @@ related utilities used by hledger commands.
 
 -}
 
-{-# LANGUAGE CPP, ScopedTypeVariables, DeriveDataTypeable, FlexibleContexts, TypeFamilies #-}
+{-# LANGUAGE CPP, ScopedTypeVariables, DeriveDataTypeable, FlexibleContexts, TypeFamilies, OverloadedStrings #-}
 
 module Hledger.Cli.CliOptions (
 
@@ -15,6 +15,8 @@ module Hledger.Cli.CliOptions (
   inputflags,
   reportflags,
   outputflags,
+  outputFormatFlag,
+  outputFileFlag,
   generalflagsgroup1,
   generalflagsgroup2,
   generalflagsgroup3,
@@ -25,6 +27,8 @@ module Hledger.Cli.CliOptions (
   argsFlag,
   showModeUsage,
   withAliases,
+  likelyExecutablesInPath,
+  hledgerExecutablesInPath,
 
   -- * CLI options
   CliOpts(..),
@@ -47,6 +51,7 @@ module Hledger.Cli.CliOptions (
   outputFormatFromOpts,
   defaultWidth,
   widthFromOpts,
+  replaceNumericFlags,
   -- | For register:
   registerWidthsFromOpts,
   maybeAccountNameDrop,
@@ -58,8 +63,12 @@ module Hledger.Cli.CliOptions (
   topicForMode,
 
   -- * Tests
-  tests_Hledger_Cli_CliOptions
+  tests_Hledger_Cli_CliOptions,
 
+--  -- * Convenience re-exports
+--  module Data.String.Here,
+--  module System.Console.CmdArgs.Explicit,
+--  module Test.HUnit
 )
 where
 
@@ -67,6 +76,7 @@ import Prelude ()
 import Prelude.Compat
 import qualified Control.Exception as C
 import Control.Monad (when)
+import Data.Char
 import Data.Default
 #if !MIN_VERSION_base(4,8,0)
 import Data.Functor.Compat ((<$>))
@@ -76,6 +86,7 @@ import Data.List.Compat
 import Data.List.Split (splitOneOf)
 import Data.Ord
 import Data.Maybe
+--import Data.String.Here
 -- import Data.Text (Text)
 import qualified Data.Text as T
 import Safe
@@ -90,7 +101,7 @@ import System.Environment
 import System.Exit (exitSuccess)
 import System.FilePath
 import Test.HUnit
-import Text.Megaparsec
+import Text.Megaparsec.Compat
 
 import Hledger
 import Hledger.Cli.DocFiles
@@ -102,10 +113,7 @@ import Hledger.Cli.Version
 -- | Common help flags: --help, --debug, --version...
 helpflags :: [Flag RawOpts]
 helpflags = [
-  flagNone ["h"]    (setboolopt "h")    "show general usage (or after CMD, command usage)"
- ,flagNone ["help"] (setboolopt "help") "show this program's manual as plain text (or after an addon CMD, the add-on's manual)"
- ,flagNone ["man"]  (setboolopt "man")  "show this program's manual with man"
- ,flagNone ["info"] (setboolopt "info") "show this program's manual with info"
+  flagNone ["help","h"] (setboolopt "help") "show general usage (or after CMD, command usage)"
  -- ,flagNone ["browse-args"] (setboolopt "browse-args") "use a web UI to select options and build up a command line"
  ,flagReq  ["debug"]    (\s opts -> Right $ setopt "debug" s opts) "[N]" "show debug output (levels 1-9, default: 1)"
  ,flagNone ["version"] (setboolopt "version") "show version information"
@@ -137,23 +145,24 @@ reportflags = [
  ,flagNone ["quarterly","Q"] (setboolopt "quarterly") "multiperiod/multicolumn report by quarter"
  ,flagNone ["yearly","Y"]    (setboolopt "yearly") "multiperiod/multicolumn report by year"
  ,flagReq  ["period","p"]    (\s opts -> Right $ setopt "period" s opts) "PERIODEXP" "set start date, end date, and/or report interval all at once (overrides the flags above)"
- ,flagNone ["date2"]         (setboolopt "date2") "show, and make -b/-e/-p/date: match, secondary dates instead"
+ ,flagNone ["date2"]         (setboolopt "date2") "match the secondary date instead (see command help for other effects)"
 
+ ,flagNone ["unmarked","U"]  (setboolopt "unmarked") "include only unmarked postings/txns (can combine with -P or -C)"
+ ,flagNone ["pending","P"]   (setboolopt "pending") "include only pending postings/txns"
  ,flagNone ["cleared","C"]   (setboolopt "cleared") "include only cleared postings/txns"
- ,flagNone ["pending"]       (setboolopt "pending") "include only pending postings/txns"
- ,flagNone ["uncleared","U"] (setboolopt "uncleared") "include only uncleared (and pending) postings/txns"
  ,flagNone ["real","R"]      (setboolopt "real") "include only non-virtual postings"
- ,flagReq  ["depth"]         (\s opts -> Right $ setopt "depth" s opts) "N" "hide accounts/postings deeper than N"
+ ,flagReq  ["depth"]         (\s opts -> Right $ setopt "depth" s opts) "NUM" "(or -NUM): hide accounts/postings deeper than this"
  ,flagNone ["empty","E"]     (setboolopt "empty") "show items with zero amount, normally hidden"
  ,flagNone ["cost","B"]      (setboolopt "cost") "convert amounts to their cost at transaction time (using the transaction price, if any)"
  ,flagNone ["value","V"]     (setboolopt "value") "convert amounts to their market value on the report end date (using the most recent applicable market price, if any)"
+ ,flagNone ["auto"]          (setboolopt "auto") "apply automated posting rules to modify transactions"
+ ,flagNone ["forecast"]      (setboolopt "forecast") "apply periodic transaction rules to generate future transactions, to 6 months from now or report end date"
  ]
 
 -- | Common output-related flags: --output-file, --output-format...
-outputflags = [
-   flagReq  ["output-format","O"] (\s opts -> Right $ setopt "output-format" s opts) "FMT" "select the output format. Supported formats:\ntxt, csv."
-  ,flagReq  ["output-file","o"]   (\s opts -> Right $ setopt "output-file" s opts) "FILE" "write output to FILE. A file extension matching one of the above formats selects that format."
-  ]
+outputflags = [outputFormatFlag, outputFileFlag]
+outputFormatFlag = flagReq  ["output-format","O"] (\s opts -> Right $ setopt "output-format" s opts) "FMT" "select the output format. Supported formats:\ntxt, csv."
+outputFileFlag   = flagReq  ["output-file","o"]   (\s opts -> Right $ setopt "output-file" s opts) "FILE" "write output to FILE. A file extension matching one of the above formats selects that format."
 
 argsFlag :: FlagHelp -> Arg RawOpts
 argsFlag desc = flagArg (\s opts -> Right $ setopt "args" s opts) desc
@@ -200,7 +209,7 @@ defCommandMode names = defMode {
   ,modeGroupFlags  = Group {
      groupNamed   = []
     ,groupUnnamed = [
-       flagNone ["h"] (setboolopt "h") "Show usage."
+       flagNone ["help"] (setboolopt "help") "Show usage."
       -- ,flagNone ["help"] (setboolopt "help") "Show long help."
       ]
     ,groupHidden  = []             --  flags not displayed in the usage
@@ -246,11 +255,10 @@ parseHelpTemplate t =
         preamble = unlines $ reverse $ dropWhile null $ reverse preamblels
         postamblelines = dropWhile null $ drop 1 postamblels
 
--- | Build a cmdarg mode suitable for a hledger add-on command,
+-- | Build a cmdarg mode for a hledger command,
 -- from a help template and flag/argument specifications.
 -- Reduces boilerplate a little, though the complicated cmdargs
 -- flag and argument specs are still required.
--- See the addons in bin/ for examples of usage.
 hledgerCommandMode :: HelpTemplate -> [Flag RawOpts] -> [(Help, [Flag RawOpts])] 
   -> [Flag RawOpts] -> ([Arg RawOpts], Maybe (Arg RawOpts)) -> Mode RawOpts
 hledgerCommandMode tmpl ungroupedflags groupedflags hiddenflags args =
@@ -336,11 +344,10 @@ data CliOpts = CliOpts {
      rawopts_         :: RawOpts
     ,command_         :: String
     ,file_            :: [FilePath]
-    ,rules_file_      :: Maybe FilePath
+    ,inputopts_       :: InputOpts
+    ,reportopts_      :: ReportOpts
     ,output_file_     :: Maybe FilePath
     ,output_format_   :: Maybe String
-    ,alias_           :: [String]
-    ,ignore_assertions_ :: Bool
     ,debug_           :: Int            -- ^ debug level, set by @--debug[=N]@. See also 'Hledger.Utils.debugLevel'.
     ,no_new_accounts_ :: Bool           -- add
     ,width_           :: Maybe String   -- ^ the --width value provided, if any
@@ -348,7 +355,6 @@ data CliOpts = CliOpts {
                                         -- 1. the COLUMNS env var, if set
                                         -- 2. the width reported by the terminal, if supported
                                         -- 3. the default (80)
-    ,reportopts_      :: ReportOpts
  } deriving (Show, Data, Typeable)
 
 instance Default CliOpts where def = defcliopts
@@ -365,9 +371,7 @@ defcliopts = CliOpts
     def
     def
     def
-    def
     defaultWidth
-    def
 
 -- | Convert possibly encoded option values to regular unicode strings.
 decodeRawOpts :: RawOpts -> RawOpts
@@ -377,12 +381,21 @@ decodeRawOpts = map (\(name',val) -> (name', fromSystemString val))
 defaultWidth :: Int
 defaultWidth = 80
 
+-- | Replace any numeric flags (eg -2) with their long form (--depth 2),
+-- as I'm guessing cmdargs doesn't support this directly.  
+replaceNumericFlags :: [String] -> [String]
+replaceNumericFlags = map replace
+  where
+    replace ('-':ds) | not (null ds) && all isDigit ds = "--depth="++ds
+    replace s = s
+
 -- | Parse raw option string values to the desired final data types.
 -- Any relative smart dates will be converted to fixed dates based on
 -- today's date. Parsing failures will raise an error.
 -- Also records the terminal width, if supported.
 rawOptsToCliOpts :: RawOpts -> IO CliOpts
 rawOptsToCliOpts rawopts = checkCliOpts <$> do
+  let iopts = rawOptsToInputOpts rawopts
   ropts <- rawOptsToReportOpts rawopts
   mcolumns <- readMay <$> getEnvSafe "COLUMNS"
   mtermwidth <-
@@ -397,16 +410,14 @@ rawOptsToCliOpts rawopts = checkCliOpts <$> do
               rawopts_         = rawopts
              ,command_         = stringopt "command" rawopts
              ,file_            = map (T.unpack . stripquotes . T.pack) $ listofstringopt "file" rawopts
-             ,rules_file_      = maybestringopt "rules-file" rawopts
+             ,inputopts_       = iopts
+             ,reportopts_      = ropts
              ,output_file_     = maybestringopt "output-file" rawopts
              ,output_format_   = maybestringopt "output-format" rawopts
-             ,alias_           = map (T.unpack . stripquotes . T.pack) $ listofstringopt "alias" rawopts
              ,debug_           = intopt "debug" rawopts
-             ,ignore_assertions_ = boolopt "ignore-assertions" rawopts
              ,no_new_accounts_ = boolopt "no-new-accounts" rawopts -- add
              ,width_           = maybestringopt "width" rawopts
              ,available_width_ = availablewidth
-             ,reportopts_      = ropts
              }
 
 -- | Do final validation of processed opts, raising an error if there is trouble.
@@ -424,6 +435,7 @@ checkCliOpts opts =
 -- and returns a CliOpts. Or, with --help or -h present, it prints 
 -- long or short help, and exits the program. 
 -- When --debug is present, also prints some debug output.
+-- Note this is not used by the main hledger executable.
 --
 -- The help texts are generated from the mode.
 -- Long help includes the full usage description generated by cmdargs
@@ -443,12 +455,12 @@ checkCliOpts opts =
 --
 getHledgerCliOpts :: Mode RawOpts -> IO CliOpts
 getHledgerCliOpts mode' = do
-  args' <- getArgs
+  args' <- getArgs >>= expandArgsAt
   let rawopts = either usageError decodeRawOpts $ process mode' args'
   opts <- rawOptsToCliOpts rawopts
   debugArgs args' opts
-  when ("help" `inRawOpts` rawopts_ opts) $ putStr longhelp  >> exitSuccess
-  when ("h"    `inRawOpts` rawopts_ opts) $ putStr shorthelp >> exitSuccess
+  when ("help" `inRawOpts` rawopts_ opts) $ putStr shorthelp >> exitSuccess
+  -- when ("help" `inRawOpts` rawopts_ opts) $ putStr longhelp  >> exitSuccess
   return opts
   where
     longhelp = showModeUsage mode'
@@ -457,7 +469,7 @@ getHledgerCliOpts mode' = do
         (reverse $ dropWhile null $ reverse $ takeWhile (not . ("flags:" `isInfixOf`)) $ lines longhelp)
         ++
         [""
-        ,"  See --help for full detail, including common hledger options."
+        ,"  See also hledger -h for general hledger options."
         ]
     -- | Print debug info about arguments and options if --debug is present.
     debugArgs :: [String] -> CliOpts -> IO ()
@@ -475,7 +487,7 @@ getHledgerCliOpts mode' = do
 -- | Get the account name aliases from options, if any.
 aliasesFromOpts :: CliOpts -> [AccountAlias]
 aliasesFromOpts = map (\a -> fromparse $ runParser accountaliasp ("--alias "++quoteIfNeeded a) $ T.pack a)
-                  . alias_
+                  . aliases_ . inputopts_
 
 -- | Get the (tilde-expanded, absolute) journal file path from
 -- 1. options, 2. an environment variable, or 3. the default.
@@ -539,7 +551,7 @@ filePathExtension = dropWhile (=='.') . snd . splitExtension . snd . splitFileNa
 rulesFilePathFromOpts :: CliOpts -> IO (Maybe FilePath)
 rulesFilePathFromOpts opts = do
   d <- getCurrentDirectory
-  maybe (return Nothing) (fmap Just . expandPath d) $ rules_file_ opts
+  maybe (return Nothing) (fmap Just . expandPath d) $ mrules_file_ $ inputopts_ opts
 
 -- | Get the width in characters to use for console output.
 -- This comes from the --width option, or the COLUMNS environment
@@ -548,7 +560,7 @@ rulesFilePathFromOpts opts = do
 widthFromOpts :: CliOpts -> Int
 widthFromOpts CliOpts{width_=Nothing, available_width_=w} = w
 widthFromOpts CliOpts{width_=Just s}  =
-    case runParser (read `fmap` some digitChar <* eof :: ParsecT Dec String Identity Int) "(unknown)" s of
+    case runParser (read `fmap` some digitChar <* eof :: ParsecT MPErr String Identity Int) "(unknown)" s of
         Left e   -> usageError $ "could not parse width option: "++show e
         Right w  -> w
 
@@ -570,7 +582,7 @@ registerWidthsFromOpts CliOpts{width_=Just s}  =
         Left e   -> usageError $ "could not parse width option: "++show e
         Right ws -> ws
     where
-        registerwidthp :: (Stream s, Char ~ Token s) => ParsecT Dec s m (Int, Maybe Int)
+        registerwidthp :: (Stream s, Char ~ Token s) => ParsecT MPErr s m (Int, Maybe Int)
         registerwidthp = do
           totalwidth <- read `fmap` some digitChar
           descwidth <- optional (char ',' >> read `fmap` some digitChar)
@@ -636,23 +648,27 @@ dropRedundantSourceVersion fs = fs
 compiledExts = ["",".com",".exe"] 
 
 
--- | Get the sorted unique filenames of all hledger-* executables in
--- the current user's PATH. Currently these are: files in any of the
--- PATH directories, named hledger-*, with either no extension (and no
--- periods in the name) or one of the addonExtensions.  Limitations:
--- we do not currently check that the file is really a file (not eg a
--- directory) or whether it has execute permission.
-hledgerExecutablesInPath :: IO [String]
-hledgerExecutablesInPath = do
+-- | Get all sorted unique filenames in the current user's PATH. 
+-- We do not currently filter out non-file objects or files without execute permission.
+likelyExecutablesInPath :: IO [String]
+likelyExecutablesInPath = do
   pathdirs <- splitOneOf "[:;]" `fmap` getEnvSafe "PATH"
   pathfiles <- concat `fmap` mapM getDirectoryContentsSafe pathdirs
-  return $ nub $ sort $ filter isHledgerExeName pathfiles
-  -- XXX should exclude directories and files without execute permission.
+  return $ nub $ sort pathfiles
+  -- exclude directories and files without execute permission.
   -- These will do a stat for each hledger-*, probably ok.
   -- But they need paths, not just filenames
-  -- hledgerexes  <- filterM doesFileExist hledgernamed
-  -- hledgerexes' <- filterM isExecutable hledgerexes
-  -- return hledgerexes
+  -- exes'  <- filterM doesFileExist exe'
+  -- exes'' <- filterM isExecutable exes'
+  -- return exes''
+
+-- | Get the sorted unique filenames of all hledger-* executables in
+-- the current user's PATH. These are files in any of the PATH directories,
+-- named hledger-*, with either no extension (and no periods in the name) 
+-- or one of the addonExtensions. 
+-- We do not currently filter out non-file objects or files without execute permission.
+hledgerExecutablesInPath :: IO [String]
+hledgerExecutablesInPath = filter isHledgerExeName <$> likelyExecutablesInPath
 
 -- isExecutable f = getPermissions f >>= (return . executable)
 
@@ -660,10 +676,10 @@ isHledgerExeName :: String -> Bool
 isHledgerExeName = isRight . parsewith hledgerexenamep . T.pack
     where
       hledgerexenamep = do
-        _ <- string progname
+        _ <- mptext $ T.pack progname
         _ <- char '-'
-        _ <- some (noneOf ".")
-        optional (string "." >> choice' (map string addonExtensions))
+        _ <- some $ noneOf ['.']
+        optional (string "." >> choice' (map (mptext . T.pack) addonExtensions))
         eof
 
 stripAddonExtension :: String -> String
